

#include	<stdio.h>
#include	<sys/types.h>
#include	<dirent.h>
#include	<unistd.h>
#include	<time.h>
#include	<errno.h>


#define		MAX_ACCT	6760
#define		ACCT_DIR	"/u/oly/act"
#define		OLD_ACCT_DIR	"/u/oly/old-act"
#define		LEN		256
#define		TRUE		1
#define		FALSE		0

void trans(char *pl, char *amount, char *comment);
void log(char *pl, double val, char *comment);
void turn(char *pl, char *comment);

double atof();

usage()
{
    fprintf(stderr, "usage: acct [options]\n");
    fprintf(stderr, "   -n    add a new player (info file on stdin)\n");
    fprintf(stderr, "   -a    email the new player a short ACK\n");
    fprintf(stderr, "   -l n  log credit/debit of n (give p and y)\n");
    fprintf(stderr, "   -t    charge a turn (give p and y)\n");
    fprintf(stderr, "   -i      this is an idle turn (reduced rate)\n");
    fprintf(stderr, "   -p n  player account to credit/debit to\n");
    fprintf(stderr, "   -y n  comment for credit/debit\n");
    fprintf(stderr, "   -s n  account summary, last n log entries\n");
    fprintf(stderr, "   -m    list accounts at or below zero\n");
    fprintf(stderr, "   -c       ... only if they pay by credit card\n");
    fprintf(stderr, "   -f    list players with free turns\n");
    fprintf(stderr, "   -A    list all accounts\n");
    fprintf(stderr, "   -d    delete player account\n");
    exit(1);
}


int new_flag = 0;	/* add a new player */
int ack_flag = 0;	/* send the new player a short acknowledgement */
int log_flag = 0;
int show_flag = 0;
int turn_flag = 0;
int idle_turn_flag = 0;
int list_flag = 0;
int list_all_flag = 0;
int cc_flag = 0;
int free_flag = 0;
int del_flag = 0;

char player[LEN];
char val[LEN];
char comment[LEN];


main(argc, argv)
int argc;
char **argv;
{
	extern int optind, opterr;
	extern char *optarg;
	int errflag = 0;
	int c;

	strcpy(comment, "n/a");
	strcpy(val, "0");
	strcpy(player, "0");

	while ((c = getopt(argc, argv, "diAfmctl:p:y:nas:?")) != EOF)
	{
		switch (c)
		{
		case 'd':
			del_flag++;
			break;

		case 'A':
			list_all_flag++;
			break;

		case 'f':
			free_flag++;
			break;

		case 't':
			turn_flag++;
			break;

		case 'i':
			idle_turn_flag++;
			break;

		case 'l':
			log_flag++;
			strcpy(val, optarg);
			break;

		case 'n':
			new_flag++;
			break;

		case 'a':
			ack_flag++;
			break;

		case 'p':
			strcpy(player, optarg);
			break;

		case 'y':
			strcpy(comment, optarg);
			break;

		case 's':
			show_flag = atoi(optarg);
			break;

		case 'm':
			list_flag++;
			break;

		case 'c':
			cc_flag++;
			break;

		case '?':
		default:
			errflag++;
			break;
		}
	}

	if (errflag)
		usage();	/* doesn't return */

	if (new_flag)
		new_account();
	else if (list_flag)
		list_cc_below_zero();
	else if (free_flag)
		list_free_players();
	else if (log_flag)
		trans(player, val, comment);
	else if (show_flag)
		show(player);
	else if (turn_flag)
		turn(player, comment);
	else if (list_all_flag)
		list_all();
	else if (del_flag)
		delete_account(player);
	else
		usage();

	exit(0);
}


char *
alloc_account()
{
	char fnam[LEN];
	char new[LEN];
	FILE *fp;
	FILE *newfp;
	char buf[LEN];
	static char acc[LEN];
	char *p;

	sprintf(fnam, "%s/.alloc", ACCT_DIR);
	sprintf(new, "%s/.new_alloc", ACCT_DIR);

	fp = fopen(fnam, "r");
	if (fp == NULL)
	{
		fprintf(stderr, "can't read %s:", fnam);
		perror("");
		exit(1);
	}

	if (fgets(acc, LEN, fp) == NULL)
	{
		fprintf(stderr, "file %s empty\n", fnam);
		exit(1);
	}

	newfp = fopen(new, "w");
	if (newfp == NULL)
	{
		fprintf(stderr, "can't write %s:", new);
		perror("");
		exit(1);
	}

	while (fgets(buf, LEN, fp) != NULL)
	{
		fputs(buf, newfp);
	}

	fclose(newfp);
	fclose(fp);

	unlink(fnam);
	if (rename(new, fnam) < 0)
	{
		fprintf(stderr, "rename(%s, %s) failed:", new, fnam);
		perror("");
		exit(1);
	}

	for (p = acc; *p && *p != '\n'; p++)
		;
	*p = '\0';

	return acc;
}



char *months[] = {
	"jan", "feb", "mar", "apr", "may", "jun",
	"jul", "aug", "sep", "oct", "nov", "dec"
};


char *
get_date() {
	struct tm *t;
	long l;
	static char buf[LEN];

	time(&l);
	t = localtime(&l);

	sprintf(buf, "%2d-%s-%2d", t->tm_mday, months[t->tm_mon], t->tm_year);

	return buf;
}


char *
get_who(char *pl)
{
	static char ret[LEN];
	char buf[LEN];
	char fnam[LEN];
	FILE *fp;
	char *p;

	*ret = '\0';

	sprintf(fnam, "%s/%s/Join", ACCT_DIR, pl);

	fp = fopen(fnam, "r");
	if (fp == NULL)
		strcat(fnam, "-");
	fp = fopen(fnam, "r");
	if (fp == NULL)
	{
		fprintf(stderr, "can't read %s:", fnam);
		perror("");
		return ret;
	}

	if (fgets(buf, LEN, fp))
	{
	  if (fgets(buf, LEN, fp))
	  {
	    if (fgets(ret, LEN, fp))
	    {
		for (p = ret; *p && *p != '\n'; p++)
			;
		*p = '\0';

		if (fgets(buf, LEN, fp))
		{
			for (p = buf; *p && *p != '\n'; p++)
				;
			*p = '\0';

			strcat(ret, " <");
			strcat(ret, buf);
			strcat(ret, ">");
		}
	    }
	  }
	}

	return ret;
}


#if 0
char *
get_who(char *pl)
{
	static char ret[LEN];
	char buf[LEN];
	char fnam[LEN];
	FILE *fp;
	char *p;

	*ret = '\0';

	sprintf(fnam, "%s/%s/Join", ACCT_DIR, pl);

	fp = fopen(fnam, "r");
	if (fp == NULL)
		strcat(fnam, "-");
	fp = fopen(fnam, "r");
	if (fp == NULL)
	{
		fprintf(stderr, "can't read %s:", fnam);
		perror("");
		return ret;
	}

	if (fgets(buf, LEN, fp))
	{
	  if (fgets(buf, LEN, fp))
	  {
	    if (fgets(ret, LEN, fp))
	    {
		for (p = ret; *p && *p != '\n'; p++)
			;
		*p = '\0';

		if (fgets(buf, LEN, fp))
		{
			for (p = buf; *p && *p != '\n'; p++)
				;
			*p = '\0';

			strcat(ret, " <");
			strcat(ret, buf);
			strcat(ret, ">");
		}
	    }
	  }
	}

	return ret;
}
#endif


void
load_acct(char *pl, double *bal, int *free, double *cash_bal)
{
	FILE *fp;
	char fnam[LEN];

	sprintf(fnam, "%s/%s/bal", ACCT_DIR, pl);

	fp = fopen(fnam, "r");
	if (fp == NULL)
	{
		fprintf(stderr, "can't read %s:", fnam);
		perror("");
		exit(1);
	}
	fscanf(fp, "%lf", bal);
	fclose(fp);

	if (cash_bal)
	{
		sprintf(fnam, "%s/.cash_bal", ACCT_DIR);

		fp = fopen(fnam, "r");
		if (fp == NULL)
		{
			fprintf(stderr, "can't read %s:", fnam);
			perror("");
			exit(1);
		}
		fscanf(fp, "%lf", cash_bal);
		fclose(fp);
	}

	if (free)
	{
		sprintf(fnam, "%s/%s/free", ACCT_DIR, pl);

		fp = fopen(fnam, "r");
		if (fp == NULL)
		{
			*free = 0;
		}
		else
		{
			fscanf(fp, "%d", free);
			fclose(fp);
		}
	}
}


void
save_acct(char *pl, double *bal, int *free, double *cash_bal)
{
	FILE *fp;
	char fnam[LEN];

	sprintf(fnam, "%s/%s/bal", ACCT_DIR, pl);

	fp = fopen(fnam, "w");
	if (fp == NULL)
	{
		fprintf(stderr, "can't write %s:", fnam);
		perror("");
		exit(1);
	}
	fprintf(fp, "%.2lf\n", *bal);
	fclose(fp);

	if (cash_bal)
	{
		sprintf(fnam, "%s/.cash_bal", ACCT_DIR);

		fp = fopen(fnam, "w");
		if (fp == NULL)
		{
			fprintf(stderr, "can't write %s:", fnam);
			perror("");
			exit(1);
		}
		fprintf(fp, "%.2lf\n", *cash_bal);
		fclose(fp);
	}

	if (free)
	{
		sprintf(fnam, "%s/%s/free", ACCT_DIR, pl);
		if (*free == 0)
		{
			if (unlink(fnam) < 0)
			{
				if (errno != ENOENT)
				{
					fprintf(stderr, "can't unlink %s:", fnam);
					perror("");
					exit(1);
				}
			}
		}
		else
		{
			fp = fopen(fnam, "w");
			if (fp == NULL)
			{
				fprintf(stderr, "can't write %s:", fnam);
				perror("");
				exit(1);
			}
			fprintf(fp, "%d\n", *free);
			fclose(fp);
		}
	}
}


void
log_cash(char *pl, double val, char *comment)
{
	FILE *fp;
	char fnam[LEN];

	sprintf(fnam, "%s/.cash_log", ACCT_DIR);

	fp = fopen(fnam, "a+");
	if (fp == NULL)
	{
		fprintf(stderr, "can't write %s:", fnam);
		perror("");
		exit(1);
	}

	fprintf(fp, "%s\t%s\t%6.2lf\t%s\n", get_date(), pl, val, comment);
	fclose(fp);
}


void
log(char *pl, double val, char *comment)
{
	FILE *fp;
	char fnam[LEN];

	sprintf(fnam, "%s/%s/log", ACCT_DIR, pl);

	fp = fopen(fnam, "a+");
	if (fp == NULL)
	{
		fprintf(stderr, "can't write %s:", fnam);
		perror("");
		exit(1);
	}

	fprintf(fp, "%s\t%-25s\t%6.2lf\n", get_date(), comment, val);
	fclose(fp);
}


void
trans(char *pl, char *amount, char *comment)
{
	double val;
	double bal;
	double cash_bal;

	load_acct(pl, &bal, NULL, &cash_bal);
	val = atof(amount);
	bal += val;
	cash_bal += val;
	save_acct(pl, &bal, NULL, &cash_bal);

	log(pl, val, comment);
	log_cash(pl, val, comment);
}


void
turn(char *pl, char *comment)
{
	double bal;
	int free;
	double val;

	load_acct(pl, &bal, &free, NULL);

	if (free > 0)
	{
		free--;
		val = 0.0;
	}
	else
	{
		if (idle_turn_flag)
			val = -0.50;
		else
			val = -2.50;
	}

	bal += val;

	save_acct(pl, &bal, &free, NULL);

	log(pl, val, comment);

/*
 *  Now we leave a little marker if this player is idle
 */

	{
		char fnam[LEN];
		FILE *fp;

		sprintf(fnam, "%s/%s/idle", ACCT_DIR, pl);
		if (idle_turn_flag)
		{
			fp = fopen(fnam, "w");
			if (fp == NULL)
			{
				fprintf(stderr, "can't write %s", fnam);
				perror("");
			}
			else
				fclose(fp);
		}
		else
			unlink(fnam);
	}
}


show(pl)
char *pl;
{
	char cmd[LEN];
	double bal;
	int free;

	printf("Account: %s\n\n", pl);
	fflush(stdout);

	sprintf(cmd, "tail -%d %s/%s/log", show_flag, ACCT_DIR, pl);
	system(cmd);

	fflush(stdout);

	load_acct(pl, &bal, &free, NULL);

	printf("\nBalance:     $%.2lf\n", bal);
	if (free > 0)
		printf("Free turns:  %d\n", free);

	fflush(stdout);
}


new_account()
{
	char *acct;
	char buf[LEN];
	char fnam[LEN];
	FILE *fp;
	char from[LEN];
	char cmd[LEN];
	int in_headers = TRUE;
	char *p;

	*from = '\0';
	acct = alloc_account();
	sprintf(fnam, "%s/%s", ACCT_DIR, acct);

	if (mkdir(fnam, 0700) < 0)
	{
		fprintf(stderr, "can't mkdir %s:", fnam);
		perror("");
		exit(1);
	}

	sprintf(fnam, "%s/%s/New", ACCT_DIR, acct);

	fp = fopen(fnam, "w");
	if (fp == NULL)
	{
		fprintf(stderr, "can't write %s:", fnam);
		perror("");
		exit(1);
	}

	while (fgets(buf, LEN, stdin) != NULL)
	{
		for (p = buf; *p && *p != '\n'; p++)
			;
		*p = '\0';

		if (*buf == '\0')
			in_headers = FALSE;
		else if (in_headers && strncmp(buf, "From: ", 6) == 0)
			strcpy(from, &buf[6]);

		fprintf(fp, "%s\n", buf);
	}

	fclose(fp);

	sprintf(fnam, "%s/%s/bal", ACCT_DIR, acct);

	fp = fopen(fnam, "w");
	if (fp == NULL)
	{
		fprintf(stderr, "can't write %s:", fnam);
		perror("");
		exit(1);
	}

	fprintf(fp, "0\n");
	fclose(fp);

	printf("Created new account %s\n", acct);

	log(acct, 0.0, "account opened");

	if (ack_flag && *from)
	{
		sprintf(cmd, "/u/oly/bin/ack-signup '%s'", from);
		system(cmd);
	}
}


list_cc_below_zero()
{
	DIR *d;
	struct dirent *e;
	char fnam[LEN];
	double bal;
	int free;
	char *idle_s;

	d = opendir(ACCT_DIR);

	if (d == NULL)
	{
		fprintf(stderr, "can't open %s: ", ACCT_DIR);
		perror("");
		exit(1);
	}

	while ((e = readdir(d)) != NULL)
	{
		if (*(e->d_name) == '.')
			continue;

		load_acct(e->d_name, &bal, &free, NULL);

		if (free > 0 || bal > 0)
			continue;

		sprintf(fnam, "%s/%s/cc", ACCT_DIR, e->d_name);

		if (cc_flag && access(fnam, R_OK) < 0)
			continue;

		sprintf(fnam, "%s/%s/idle", ACCT_DIR, e->d_name);
		if (access(fnam, R_OK) < 0)
			idle_s = "   ";
		else
			idle_s = "[i]";

		printf("%s  %6.2lf %s    %s\n", e->d_name, bal, idle_s,
					get_who(e->d_name));
	}

	closedir(d);
}


list_free_players()
{
	DIR *d;
	struct dirent *e;
	char fnam[LEN];
	double bal;
	int free;

	d = opendir(ACCT_DIR);

	if (d == NULL)
	{
		fprintf(stderr, "can't open %s: ", ACCT_DIR);
		perror("");
		exit(1);
	}

	while ((e = readdir(d)) != NULL)
	{
		if (*(e->d_name) == '.')
			continue;

		load_acct(e->d_name, &bal, &free, NULL);

		if (free <= 0)
			continue;

		printf("%s  %4d	%s\n", e->d_name, free,
					get_who(e->d_name));
	}

	closedir(d);
}


list_all()
{
	DIR *d;
	struct dirent *e;
	char fnam[LEN];
	double bal;
	int free;

	d = opendir(ACCT_DIR);

	if (d == NULL)
	{
		fprintf(stderr, "can't open %s: ", ACCT_DIR);
		perror("");
		exit(1);
	}

	while ((e = readdir(d)) != NULL)
	{
		if (*(e->d_name) == '.')
			continue;

		load_acct(e->d_name, &bal, &free, NULL);

		printf("%s  %6.2lf  %4d\t%s\n", e->d_name, bal, free,
					get_who(e->d_name));
	}

	closedir(d);
}


delete_account(pl)
char *pl;
{
	char one[LEN], two[LEN];

	log(pl, 0.0, "player drop");

	mkdir(OLD_ACCT_DIR, 0700);

	sprintf(one, "%s/%s", ACCT_DIR, pl);
	sprintf(two, "%s/%s", OLD_ACCT_DIR, pl);

	if (rename(one, two) < 0)
	{
		fprintf(stderr, "rename(%s, %s) failed:", one, two);
		perror("");
		exit(1);
	}
}

